///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// File:           CC121.cpp                                                 //
//                                                                           //
// Functionality:  Control surface class implemention.                       //
//                                                                           //
// Author:         Rolf Meyerhoff (rm@matrix44.de)                           //
//                                                                           //
// Copyright:      (C) 2009 by Rolf Meyerhoff. All rights reserved.          //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// License:                                                                  //
//                                                                           //
// This library is free software;  you can redistribute it  and/or modify it //
// under the terms of the GNU Lesser  General Public License as published by //
// the Free Software Foundation;  either version 2.1 of the License,  or (at //
// your option) any later version.                                           //
//                                                                           //
// This  library is  distributed in  the hope that  it will  be useful,  but //
// WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty   of //
// MERCHANTABILITY or FITNESS FOR A  PARTICULAR PURPOSE.  See the GNU Lesser //
// General Public License for more details.                                  //
//                                                                           //
// You should have received a copy of  the GNU Lesser General Public License //
// along with this library;  if not,  write to the Free Software Foundation, //
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.              //
///////////////////////////////////////////////////////////////////////////////
#include "reaper_cc121.h"

///////////////////////////////////////////////////////////////////////////////
// CC121::CC121()
///////////////////////////////////////////////////////////////////////////////
// Description:  Default constructor of this class.
// Parameters:   None.
// Return value: No return value.
// Remarks:      Explicit construction is required so this constructor will
//               throw an exception.
///////////////////////////////////////////////////////////////////////////////
CC121::CC121()
{
  // Should never reach here.
  assert(FALSE);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::~CC121()
///////////////////////////////////////////////////////////////////////////////
// Description:  Destructor of this class.
// Parameters:   None.
// Return value: No return value.
// Remarks:      Closes the MIDI input and MIDI output and frees all
//               allocated resources.
///////////////////////////////////////////////////////////////////////////////
CC121::~CC121()
{
  // Reset surface:
  UpdateDevice(0, 0x3200);

  // Give the MIDI out enough time to process the messages:
  Sleep(5);

  // Delete MIDI in and MIDI out:
  // TODO: Both were created by REAPER through CreateMIDI*put(). Delete them
  // with the runtime delete of this plugin might have serious side effects:
  if (midiOut)
    delete midiOut;
  if (midiIn)
    delete midiIn;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::GetTypeString()
///////////////////////////////////////////////////////////////////////////////
// Description:  Retrieve the type ID of this control surface.
// Parameters:   None.
// Return value: Returns the type ID as string.
// Remarks:      This ID seems to be used to select the right surface and to
//               load and store configurations. This has to be a simple
//               unique string with only A-Z, 0-9, no spaces or other chars.
///////////////////////////////////////////////////////////////////////////////
const char* CC121::GetTypeString()
{
  // Return ID:
  return "CC121";
}

///////////////////////////////////////////////////////////////////////////////
// CC121::GetDescString()
///////////////////////////////////////////////////////////////////////////////
// Description:  Retrieve the human readable description of this surface.
// Parameters:   None.
// Return value: Returns the description as string.
// Remarks:      This description is displayed in the control surface dialog
//               box. Instance specific informations are allowed here.
///////////////////////////////////////////////////////////////////////////////
const char* CC121::GetDescString()
{
  // Temporary space for the plugin description:
  static char desctmp[1024];

  // Get human readable name of the MIDI input:
  char in_buff[256];
  if (options.midiInDev >= 0 && options.midiInDev < GetNumMIDIInputs())
    GetMIDIInputName(options.midiInDev, in_buff, sizeof(in_buff) / sizeof(in_buff[0]));
  else
    strcpy(in_buff, "No Input");
   
  // Get human readable name of the MIDI output:
  char out_buff[256];
  if (options.midiOutDev >= 0 && options.midiOutDev < GetNumMIDIOutputs())
    GetMIDIOutputName(options.midiOutDev, out_buff, sizeof(out_buff) / sizeof(out_buff[0]));
  else
    strcpy(out_buff, "No Output");

  // Format display string:
  sprintf(desctmp, "Steinberg CC121 (%s, %s)", in_buff, out_buff);
  return desctmp;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::GetConfigString()
///////////////////////////////////////////////////////////////////////////////
// Description:  Retrieve the configuration of this surface.
// Parameters:   None.
// Return value: Returns the current configuration as string.
// Remarks:      This configuration string is actually stored in REAPER's
//               configuration.
///////////////////////////////////////////////////////////////////////////////
const char* CC121::GetConfigString()
{
  // Temporary space for the plugin configuration:
  static char configtmp[1024];

  // Write config into a temporary string:
  options.WriteConfig(configtmp);

  // Return the temp string:
  return configtmp;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::CloseNoReset()
///////////////////////////////////////////////////////////////////////////////
// Description:  Close this surface without resetting the state.
// Parameters:   None.
// Return value: No return value.
// Remarks:      Close without sending "reset" messages, prevent "reset"
//               being sent on destructor. 
///////////////////////////////////////////////////////////////////////////////
void CC121::CloseNoReset()
{
  // Delete MIDI in and MIDI out:
  // TODO: Both were created by REAPER through CreateMIDI*put(). Delete them
  // with the runtime delete of this plugin might have serious side effects:
  if (midiOut)
    delete midiOut;
  midiOut = NULL;
  if (midiIn)
    delete midiIn;
  midiIn = NULL;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::Run()
///////////////////////////////////////////////////////////////////////////////
// Description:  The main message handling routine.
// Parameters:   None.
// Return value: No return value.
// Remarks:      This one is called 30 times a sec or so.
///////////////////////////////////////////////////////////////////////////////
void CC121::Run()
{
  // Hardware available?
  if (midiOut == NULL || midiIn == NULL)
    return;

  // Get current time of this run:
  DWORD curTime = timeGetTime();

  // Time to send running status?
  if (curTime > (lastAISendTime + 900))
  {
    // Send message:
    SendRunningStatus();

    // Lost connection?
    if (curTime > (lastAISendTime + 2500))
    {
      // Cache old values:
      int oldLight = lightState;
      int oldFader = faderState;

      // Update state:
      lightState = 0;
      faderState = 0;
      UpdateDevice(oldLight, oldFader);
    }

    // Save send time:
    lastAISendTime = curTime;
  }

  // Time to check monitor button status?
  if (curTime > (lastMonChkTime + 100))
  {
    // Get track:
    MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
    if (tr)
    {
      // Get current state:
      int state = *static_cast<int*>(GetSetMediaTrackInfo(tr, "I_RECMON", NULL));

      // Update light state:
      SwitchLight(LightMonitor, state != 0);
    }

    // Save send time:
    lastMonChkTime = curTime;
  }

  // Time to check EQ status?
  if (curTime > (lastEQChkTime + 200))
  {
    bool state = false;
    MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
    if (tr)
    {
      // Check EQ:
      int fxcnt = TrackFX_GetCount(tr);
      if (fxcnt)
      {
        for (int fxid = 0; fxid < fxcnt; fxid++)
        {
          char fxname[256];
          if (TrackFX_GetFXName(tr, fxid, fxname, sizeof(fxname) / sizeof(fxname[0])))
          {
            if (strcmp(fxname, "VST: ReaEQ (Cockos)") == 0)
            {
              int paramcnt = TrackFX_GetNumParams(tr, fxid);
              for (int paramid = 0; paramid < paramcnt; paramid++)
              {
                char paramname[256];
                if (TrackFX_GetParamName(tr, fxid, paramid, paramname, sizeof(paramname) / sizeof(paramname[0])))
                {
                  if (strcmp(paramname, "Bypass") == 0)
                  {
                    double minval, maxval, val;
                    val = TrackFX_GetParam(tr, fxid, paramid, &minval, &maxval);
                    state = val > 0.5;
                  }
                }
              }
            }
          }
        }
      }
    }

    // Update light:
    SwitchLight(LightEQBypass, state);

    // Save send time:
    lastEQChkTime = curTime;
  }

  // Time to check for fx display?
  if (curTime > (lastFXChkTime + 200))
  {
    // Defaults to off:
    bool fxWndState = false;

    // Get track:
    MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
    if (tr)
    {
      // Format window title:
      const char* track_name = GetTrackInfo(currentTrackID - 1, NULL);
      char wnd_title[256];
      if (!track_name || track_name[0] == 0)
        sprintf(wnd_title, "FX: track %d ", currentTrackID);
      else
        sprintf(wnd_title, "FX: track %d \"%s\"", currentTrackID, track_name);
      
      // Search for the FX window:
      HWND hFx = FindWindowEx(NULL, NULL, MAKEINTATOM(32770), wnd_title);

      // Update surface if it was found:
      if (hFx != NULL && IsWindowVisible(hFx))
      {
        fxWndState = true;
      }
    }

    // Update light:
    SwitchLight(LightVSTi, fxWndState);

    // Save send time:
    lastFXChkTime = curTime;
  }

  // Time to blink with the play button?
  if (paused && (curTime > (lastPauseTime + 500)))
  {
    // Set new:
    SwitchLight(LightPlay, (lightState & LightPlay) == 0);

    // Save send time:
    lastPauseTime = curTime;
  }

  // Read MIDI events, swap to active buffer:
  midiIn->SwapBufs(curTime);

  // Get event list:
  MIDI_eventlist* list = midiIn->GetReadBuf();

  // Loop through events:
  int l = 0;
  MIDI_event_t* evts = list->EnumItems(&l);
  while (evts != NULL)
  {
    // Handle event:
    OnMIDIEvent(evts);

    // Next event:
    evts = list->EnumItems(&l);
  }

  // Seeking through project?
  if (forwarding || rewinding)
  {
    // seek thru project:
    if (forwarding)
      CSurf_OnFwd(0);
    else if (rewinding)
      CSurf_OnRew(0);
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetTrackListChange()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the track list has changed.
// Parameters:   None.
// Return value: No return value.
// Remarks:      This surface has only one channel so we don't care.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetTrackListChange()
{
  // Reset automation mode:
  SetAutoMode(0);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetSurfaceVolume()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the volume of a track has changed.
// Parameters:   [in] trackid: The track changed.
//               [in] volume:  The new volume.
// Return value: No return value.
// Remarks:      Updates the fader of the surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetSurfaceVolume(MediaTrack* trackid, double volume)
{
  // Don't update fader when touching the handle:
  if (touching)
    return;

  // Is this our selected track?
  if (CSurf_TrackToID(trackid, options.useMCP) == currentTrackID)
  {
    // Update fader state:
    if (midiOut != NULL)
    {
      // Convert from linear volume to 14 bit pitch bend:
      int volint = volToInt14(volume);

      // Update controller:
      UpdateDevice(lightState, volint);
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetSurfacePan()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the pan of a track has changed.
// Parameters:   [in] trackid: The track changed.
//               [in] pan:     The new pan.
// Return value: No return value.
// Remarks:      Does nothing as there is no fader for pan on the surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetSurfacePan(MediaTrack* trackid, double pan)
{
  // Don't care.
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetSurfaceMute()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the mute state of a track has changed.
// Parameters:   [in] trackid: The track changed.
//               [in] mute:    New mute state of the track.
// Return value: No return value.
// Remarks:      Just updates the light on the surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetSurfaceMute(MediaTrack* trackid, bool mute)
{
  // Is this our selected track?
  if (CSurf_TrackToID(trackid, options.useMCP) == currentTrackID)
  {
    // Update button state:
    SwitchLight(LightMute, mute);
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetSurfaceSelected()
///////////////////////////////////////////////////////////////////////////////
// Description:  Sets the selection state of a track.
// Parameters:   [in] trackid:  The track changed.
//               [in] selected: New selection state of the track.
// Return value: No return value.
// Remarks:      Validates the selection state of this surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetSurfaceSelected(MediaTrack* trackid, bool selected)
{
  // First time selection?
  if (selected && (currentTrackID == -1))
  {
    // Initialize currently selected track:
    currentTrackID = CSurf_TrackToID(trackid, options.useMCP);
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetSurfaceSolo()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the solo state of a track has changed.
// Parameters:   [in] trackid: The track changed.
//               [in] solo:    New solo state of the track.
// Return value: No return value.
// Remarks:      Just updates the light on the surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetSurfaceSolo(MediaTrack* trackid, bool solo)
{
  // Is this our selected track?
  if (CSurf_TrackToID(trackid, options.useMCP) == currentTrackID)
  {
    // Update button state:
    SwitchLight(LightSolo, solo);
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetSurfaceRecArm()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the record arm of a track has changed.
// Parameters:   [in] trackid: The track changed.
//               [in] recarm:  New record arm state of the track.
// Return value: No return value.
// Remarks:      Just updates the light on the surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetSurfaceRecArm(MediaTrack* trackid, bool recarm)
{
  // Is this our selected track?
  if (CSurf_TrackToID(trackid, options.useMCP) == currentTrackID)
  {
    // Update button state:
    SwitchLight(LightRecArm, recarm);
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetPlayState()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification about a change in transport state.
// Parameters:   [in] play:  Current state of the play button.
//               [in] pause: Current state of the pause button.
//               [in] rec:   Current state of the record button.
// Return value: No return value.
// Remarks:      Updates the corresponding lights on the surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetPlayState(bool play, bool pause, bool rec)
{
  // Update play and record buttons:
  SwitchLight(LightPlay, play || pause);
  SwitchLight(LightRec, rec);

  // Store paused state:
  paused = pause;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetRepeatState()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the repeat state has changed.
// Parameters:   [in] bool: New repeat state.
// Return value: No return value.
// Remarks:      Just updates the light on the surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetRepeatState(bool rep)
{
  // Update button:
  SwitchLight(LightLoop, rep);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetTrackTitle()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the name of a track has changed.
// Parameters:   [in] trackid: The track changed.
//               [in] title:   New title of the track.
// Return value: No return value.
// Remarks:      We do nothing as we don't have a place to display this info.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetTrackTitle(MediaTrack* trackid, const char* title)
{
  // Don't care.
}

///////////////////////////////////////////////////////////////////////////////
// CC121::GetTouchState()
///////////////////////////////////////////////////////////////////////////////
// Description:  Request for the current touch states of the faders.
// Parameters:   [in] trackid: The track in question.
//               [in] isPan:   Check pan fader?
// Return value: Returns true if the fader currently has been touched.
// Remarks:      This is used for automation to detect fader moves. Only the
//               volume fader is supported now.
///////////////////////////////////////////////////////////////////////////////
bool CC121::GetTouchState(MediaTrack* trackid, int isPan)
{
  // Is this our selected track?
  if (CSurf_TrackToID(trackid, options.useMCP) == currentTrackID)
    return false;

  // Our panner has no touch mode:
  if (isPan)
    return false;

  // Return current fader handle state:
  return touching;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetRepeatState()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notification that the automation mode has changed.
// Parameters:   [in] mode: New automation.
// Return value: No return value.
// Remarks:      Uses the currently selected track as this parameter seems to
//               be missing here. Just updates the automation lights on the
//               surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetAutoMode(int mode)
{
  // Hardware available?
  if (midiOut)
  {
    // Get current track and it's automation mode:
    MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
    if (tr)
      mode = GetTrackAutomationMode(tr);

    // Trim/Read mode, both lights off:
    if (mode == 0)
    {
      SwitchLight(LightRead,  false);
      SwitchLight(LightWrite, false);
    }

    // Read mode, R is on, W is off:
    else if (mode == 1)
    {
      SwitchLight(LightRead,  true);
      SwitchLight(LightWrite, false);
    }

    // Touch and Latch mode, both lights on:
    else if (mode == 2 || mode == 4)
    {
      SwitchLight(LightRead,  true);
      SwitchLight(LightWrite, true);
    }

    // Write mode, R is off, W is on:
    else if (mode == 3)
    {
      SwitchLight(LightRead,  false);
      SwitchLight(LightWrite, true);
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::ResetCachedVolPanStates()
///////////////////////////////////////////////////////////////////////////////
// Description:  Request to reset internal control states.
// Parameters:   None.
// Return value: No return value.
// Remarks:      Good to flush your control states.
///////////////////////////////////////////////////////////////////////////////
void CC121::ResetCachedVolPanStates()
{
  // Reset states:
  faderState = 0x3200;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::OnTrackSelection()
///////////////////////////////////////////////////////////////////////////////
// Description:  Notify the extension about newly selected track.
// Parameters:   None.
// Return value: No return value.
// Remarks:      Updates the internal selection states.
///////////////////////////////////////////////////////////////////////////////
void CC121::OnTrackSelection(MediaTrack* trackid)
{
  // Get track ID:
  int newpos = CSurf_TrackToID(trackid, options.useMCP);

  // Is this already our track?
  if ((newpos >= 0) && (newpos != currentTrackID))
  {
    // Update internal track state:
    SkipTracks(newpos - currentTrackID, false);

    // Update control surface:
    TrackList_UpdateAllExternalSurfaces();
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::IsKeyDown()
///////////////////////////////////////////////////////////////////////////////
// Description:  Request for the state of the special keys on a surface.
// Parameters:   [in] key: The key in question.
// Return value: Returns true if the key is down or false otherwise.
// Remarks:      The key may be VK_CONTROL, VK_MENU, VK_SHIFT, etc, whatever
//               makes sense for a surface.
///////////////////////////////////////////////////////////////////////////////
bool CC121::IsKeyDown(int key)
{
  // We don't have special keys:
  return false;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::Extended()
///////////////////////////////////////////////////////////////////////////////
// Description:  Extended command handling.
// Parameters:   [in]     call:  Message ID.
//               [in/out] parm1: First message parameter.
//               [in/out] parm2: Second message parameter.
//               [in/out] parm3: Third message parameter.
// Return value: Returns a value depending on the call ID. Returning zero
//               means that the call is not supported.
// Remarks:      There are no public defined message IDs yet so this function
//               always returns zero.
///////////////////////////////////////////////////////////////////////////////
int CC121::Extended(int call, void* parm1, void* parm2, void* parm3)
{
  // Unsupported:
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::CreateInstance()
///////////////////////////////////////////////////////////////////////////////
// Description:  Create an instance of this plugin.
// Parameters:   [in]  typeString:   Denotes the type of plugin to create (in
//                                   case an extension has registered more
//                                   plugins with this creation function).
//               [in]  configString: Configuration to use as string.
//               [out] errStats:     Returns error statistics back to REAPER.
//                                   Bit 0 denotes an input error and bit 1
//                                   an output error.
// Return value: Returns the newly created instance.
// Remarks:      This function will be registered to REAPER when the plugin
//               gets loaded. REAPER calls this function whenever it needs an
//               instance of this surface.
///////////////////////////////////////////////////////////////////////////////
IReaperControlSurface* CC121::CreateInstance(const char* typeString, const char* configString, int* errStats)
{
  // Create control surface class:
  // TODO: The created class will be deleted by REAPER's runtime. This might
  // intruduce serious side effects:
  return new CC121(CC121Options(configString), errStats);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::CC121()
///////////////////////////////////////////////////////////////////////////////
// Description:  Copy constructor of this class.
// Parameters:   [in] other: The source instance to copy.
// Return value: No return value.
// Remarks:      Copying this class is prohibited and explicit construction
//               is required so this constructor is declared as private.
///////////////////////////////////////////////////////////////////////////////
CC121::CC121(const CC121& other)
{
  // Should never reach here.
  assert(FALSE);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::CC121()
///////////////////////////////////////////////////////////////////////////////
// Description:  Initialization constructor of this class.
// Parameters:   [in]  opt:      Options for this control surface.
//               [out] errStats: Returns error stats to REAPER.
// Return value: No return value.
// Remarks:      Initializes the members and opens the required MIDI input
//               and MIDI output.
///////////////////////////////////////////////////////////////////////////////
CC121::CC121(const CC121Options& opt, int* errStats) :
  options(opt),
  lastAISendTime(0),
  midiIn(NULL),
  midiOut(NULL),
  rewinding(false),
  forwarding(false),
  jogDown(false),
  jogSpeed(2.0),
  touching(false),
  lastMonChkTime(0),
  lastEQChkTime(0),
  lastFXChkTime(0),
  currentTrackID(-1),
  paused(false),
  lastPauseTime(0),
  lightState(0),
  faderState(0x3200)
{
  // Create MIDI input:
  if (options.midiInDev >= 0)
  {
    midiIn = CreateMIDIInput(options.midiInDev);
    if (midiIn == NULL)
    {
      options.midiInDev = -1;
      if (errStats != NULL)
        *errStats |= 1;
    }
  }

  // Create MIDI output:
  if (options.midiOutDev >= 0)
  {
    midiOut = CreateThreadedMIDIOutput(CreateMIDIOutput(options.midiOutDev, false, NULL));
    if (midiOut == NULL)
    {
      options.midiOutDev = -1;
      if (errStats != NULL)
        *errStats |= 2;
    }
  }

  // Start the wait for MIDI events:
  if (midiIn)
    midiIn->start();

  // Reset surface:
  UpdateDevice(0, 0x3200);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::operator = ()
///////////////////////////////////////////////////////////////////////////////
// Description:  Assignment operator of this class.
// Parameters:   [in] other: The source instance to copy.
// Return value: No return value.
// Remarks:      Copying this class is prohibited and explicit construction
//               is required so this operator is declared as private too.
///////////////////////////////////////////////////////////////////////////////
void CC121::operator = (const CC121& other)
{
  // Should never reach here.
  assert(FALSE);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::UpdateDevice()
///////////////////////////////////////////////////////////////////////////////
// Description:  Internal helper function to update the status of the device.
// Parameters:   [in] lts: New state of the lights (see enum above).
//               [in] vol: 
// Return value: No return value.
// Remarks:      This function is all that is needed to set the state of the
//               surface.
///////////////////////////////////////////////////////////////////////////////
void CC121::UpdateDevice(const int lts, const int vol)
{
  // Anything to do?
  if (lts == lightState && vol == faderState)
    return;

  if (midiOut)
  {
    // Update fader:
    if (vol != faderState)
      midiOut->Send(0xE0, vol & 0x7F, (vol >> 7) & 0x7F, -1);

    // Get bitwise difference of the current and the new state. This gives us
    // all the lights that have different settings and need an update:
    int lightDiff = lts ^ lightState;

    // Anything to change?
    if (lightDiff)
    {
      #define LIGHT_NOTE_ON(_light, _note) \
      if (lightDiff & _light) \
        midiOut->Send(0x90, _note, (lts & _light) ? 0x7F : 0x00, -1);

      #define LIGHT_NOTE_OFF(_light, _note) \
      if (lightDiff & _light) \
        midiOut->Send(0x80, _note, 0x00, -1);

      // Note-on messages:
      LIGHT_NOTE_ON(LightRecArm,   0x00);
      LIGHT_NOTE_ON(LightSolo,     0x08);
      LIGHT_NOTE_ON(LightMute,     0x10);
      LIGHT_NOTE_ON(LightEdit,     0x33);
      LIGHT_NOTE_ON(LightUser1,    0x36);
      LIGHT_NOTE_ON(LightUser2,    0x37);
      LIGHT_NOTE_ON(LightUser3,    0x38);
      LIGHT_NOTE_ON(LightUser4,    0x39);
      LIGHT_NOTE_ON(LightRead,     0x4A);
      LIGHT_NOTE_ON(LightWrite,    0x4B);
      LIGHT_NOTE_ON(LightLoop,     0x56);
      LIGHT_NOTE_ON(LightPlay,     0x5E);
      LIGHT_NOTE_ON(LightRec,      0x5F);
      LIGHT_NOTE_ON(LightEQ1,      0x70);
      LIGHT_NOTE_ON(LightEQ2,      0x71);
      LIGHT_NOTE_ON(LightEQ3,      0x72);
      LIGHT_NOTE_ON(LightEQ4,      0x73);
      LIGHT_NOTE_ON(LightEQType,   0x74);
      LIGHT_NOTE_ON(LightEQBypass, 0x75);
      LIGHT_NOTE_ON(LightJog,      0x76);
      LIGHT_NOTE_ON(LightLock,     0x77);
      LIGHT_NOTE_ON(LightMonitor,  0x78);
      LIGHT_NOTE_ON(LightVSTi,     0x79);

      // Add a small pause to give the surface a break to handle all
      // the note-ons:
      Sleep(5);

      // Note-off messages:
      LIGHT_NOTE_OFF(LightRecArm,   0x00);
      LIGHT_NOTE_OFF(LightSolo,     0x08);
      LIGHT_NOTE_OFF(LightMute,     0x10);
      LIGHT_NOTE_OFF(LightEdit,     0x33);
      LIGHT_NOTE_OFF(LightUser1,    0x36);
      LIGHT_NOTE_OFF(LightUser2,    0x37);
      LIGHT_NOTE_OFF(LightUser3,    0x38);
      LIGHT_NOTE_OFF(LightUser4,    0x39);
      LIGHT_NOTE_OFF(LightRead,     0x4A);
      LIGHT_NOTE_OFF(LightWrite,    0x4B);
      LIGHT_NOTE_OFF(LightLoop,     0x56);
      LIGHT_NOTE_OFF(LightPlay,     0x5E);
      LIGHT_NOTE_OFF(LightRec,      0x5F);
      LIGHT_NOTE_OFF(LightEQ1,      0x70);
      LIGHT_NOTE_OFF(LightEQ2,      0x71);
      LIGHT_NOTE_OFF(LightEQ3,      0x72);
      LIGHT_NOTE_OFF(LightEQ4,      0x73);
      LIGHT_NOTE_OFF(LightEQType,   0x74);
      LIGHT_NOTE_OFF(LightEQBypass, 0x75);
      LIGHT_NOTE_OFF(LightJog,      0x76);
      LIGHT_NOTE_OFF(LightLock,     0x77);
      LIGHT_NOTE_OFF(LightMonitor,  0x78);
      LIGHT_NOTE_OFF(LightVSTi,     0x79);
    }
  }

  // Save state:
  lightState = lts;
  faderState = vol;
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SwitchLight()
///////////////////////////////////////////////////////////////////////////////
// Description:  Set the state of a specific light on the control surface.
// Parameters:   [in] which: Light to toggle (for values see enum above).
//               [in] on:    Turn this light on or off?
// Return value: No return value.
// Remarks:      This is just a shortcut to the UpdateDevice function above.
//               The state of the light is changed immediately.
///////////////////////////////////////////////////////////////////////////////
void CC121::SwitchLight(int which, bool on)
{
  // Choose flags:
  if (on)
    UpdateDevice(lightState | which, faderState);
  else
    UpdateDevice(lightState & ~which, faderState);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SendRunningStatus()
///////////////////////////////////////////////////////////////////////////////
// Description:  Sends a message to the surface to ensure 'Cubase Ready'.
// Parameters:   None.
// Return value: No return value.
// Remarks:      The CC121 will need to receive this message once a second to
//               remain operational. If the message has not been sent for
//               about three seconds then the device will turn off the lights
//               and the 'Cubase Ready' LED starts to blink.
///////////////////////////////////////////////////////////////////////////////
void CC121::SendRunningStatus()
{
  // Hardware available?
  if (midiOut != NULL)
  {
    // Create running status SysEx string:
    struct
    {
      MIDI_event_t  evt;
      unsigned char data[8];
    } sysex;
    sysex.evt.frame_offset    = 0;
    sysex.evt.size            = 8;
    sysex.evt.midi_message[0] = 0xF0;
    sysex.evt.midi_message[1] = 0x43;
    sysex.evt.midi_message[2] = 0x10;
    sysex.evt.midi_message[3] = 0x3E;
    sysex.evt.midi_message[4] = 0x15;
    sysex.evt.midi_message[5] = 0x00;
    sysex.evt.midi_message[6] = 0x01;
    sysex.evt.midi_message[7] = 0xF7;

    // Send the message:
    Sleep(5);
    midiOut->SendMsg(&sysex.evt, -1);
    Sleep(5);
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::()
///////////////////////////////////////////////////////////////////////////////
// Description:  Internal helper function to handle incomming MIDI events.
// Parameters:   [in] evt: The event to decode.
// Return value: No return value.
// Remarks:      This is the function where the actual messages of the
//               hardware are decoded and dispatched.
///////////////////////////////////////////////////////////////////////////////
void CC121::OnMIDIEvent(MIDI_event_t* evt)
{
  // Note on?
  if (evt->midi_message[0] == 0x90 && evt->midi_message[2] > 0)
  {
    // Start button?
    if (evt->midi_message[1] == 0x5E)
      CSurf_OnPlay();

    // Stop button?
    else if (evt->midi_message[1] == 0x5D)
      CSurf_OnStop();

    // Record button?
    else if (evt->midi_message[1] == 0x5F)
      CSurf_OnRecord();

    // Loop button?
    else if (evt->midi_message[1] == 0x56)
      Main_OnCommand(IDC_REPEAT, 0);

    // Skip Left button?
    else if (evt->midi_message[1] == 0x58)
    {
      if (options.skipToStartEnd)
        CSurf_GoStart();
      else
        Main_OnCommand(ID_MARKER_PREV, 0);
    }

    // Skip Left button?
    else if (evt->midi_message[1] == 0x5A)
    {
      if (options.skipToStartEnd)
        CSurf_GoEnd();
      else
        Main_OnCommand(ID_MARKER_NEXT, 0);
    }

    // Forward button?
    else if (evt->midi_message[1] == 0x5C && !rewinding)
    {
      forwarding   = true;
    }

    // Rewind button?
    else if (evt->midi_message[1] == 0x5B && !forwarding)
    {
      rewinding    = true;
    }

    // Jog button?
    else if (evt->midi_message[1] == 0x76)
    {
      // Set state
      jogDown = !jogDown;

      // Set light:
      SwitchLight(LightJog, jogDown);
    }

    // Mute button?
    else if (evt->midi_message[1] == 0x10)
    {
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
        SetSurfaceMute(tr, CSurf_OnMuteChange(tr, -1));
    }

    // Solo button?
    else if (evt->midi_message[1] == 0x08)
    {
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
        SetSurfaceSolo(tr, CSurf_OnSoloChange(tr, -1));
    }

    // Record arm button?
    else if (evt->midi_message[1] == 0x00)
    {
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
        SetSurfaceRecArm(tr, CSurf_OnRecArmChange(tr, -1));
    }

    // Prev channel button?
    else if (evt->midi_message[1] == 0x30)
    {
      SkipTracks(-1, true);
      TrackList_UpdateAllExternalSurfaces();
    }

    // Next channel button?
    else if (evt->midi_message[1] == 0x31)
    {
      SkipTracks(1, true);
      TrackList_UpdateAllExternalSurfaces();
    }

    // Automation read?
    else if (evt->midi_message[1] == 0x4A)
    {
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
      {
        // Get current automation mode:
        int mode = GetTrackAutomationMode(tr);

        // Current mode is Trim/Read, switch to Read:
        if (mode == 0)
          SetTrackAutomationMode(tr, 1);

        // Current mode is Read, switch to Trim/Read:
        else if (mode == 1)
          SetTrackAutomationMode(tr, 0);

        // Current mode is Touch, switch to Write:
        else if (mode == 2)
          SetTrackAutomationMode(tr, 4);

        // Current mode is Write, switch to Touch:
        else if (mode == 4)
          SetTrackAutomationMode(tr, 2);

        // Ignore Latch mode:
        else
          return;

        // Update surface:
        CSurf_SetAutoMode(-1, NULL);
      }
    }

    // Automation write?
    else if (evt->midi_message[1] == 0x4B)
    {
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
      {
        // Get current automation mode:
        int mode = GetTrackAutomationMode(tr);

        // Current mode is Trim/Read, switch to Write:
        if (mode == 0)
          SetTrackAutomationMode(tr, 3);

        // Current mode is Read, switch to Touch:
        else if (mode == 1)
          SetTrackAutomationMode(tr, 2);

        // Current mode is Touch, switch to Read:
        else if (mode == 2)
          SetTrackAutomationMode(tr, 1);

        // Current mode is Write, switch to Trim/Read:
        else if (mode == 3)
          SetTrackAutomationMode(tr, 0);

        // Ignore Latch mode:
        else
          return;

        // Update surface:
        CSurf_SetAutoMode(-1, NULL);
      }
    }

    // Touched fader?
    else if (evt->midi_message[1] == 0x68)
      touching = true;

    // Monitor?
    else if (evt->midi_message[1] == 0x78)
    {
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
      {
        // Get current state:
        int state = *static_cast<int*>(GetSetMediaTrackInfo(tr, "I_RECMON", NULL));

        // Set new state:
        state = (state + 1) % 3;
        GetSetMediaTrackInfo(tr, "I_RECMON", &state);

        // Update light state:
        SwitchLight(LightMonitor, state != 0);

        // Init timer:
        lastMonChkTime = timeGetTime();
      }
    }

    // 'e' Button?
    else if (evt->midi_message[1] == 0x33)
    {
      // Toggle docker:
      Main_OnCommand(ID_SHOW_DOCKER, 0);
    }

    // Value Button?
    else if (evt->midi_message[1] == 0x3A)
    {
      // Toggle click:
      Main_OnCommand(ID_TOGGLE_METRONOME, 0);
    }

    // VSTi Button?
    else if (evt->midi_message[1] == 0x79)
    {
      // Get track:
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
      {
        // Format window title:
        const char* track_name = GetTrackInfo(currentTrackID - 1, NULL);
        char wnd_title[256];
        if (!track_name || track_name[0] == 0)
          sprintf(wnd_title, "FX: track %d ", currentTrackID);
        else
          sprintf(wnd_title, "FX: track %d \"%s\"", currentTrackID, track_name);
        
        // Search for the FX window:
        HWND hFx = FindWindowEx(NULL, NULL, MAKEINTATOM(32770), wnd_title);

        // FX editor open?
        if (hFx != NULL)
        {
          // Found fx window, close it:
          SendMessage(hFx, WM_CLOSE, 0, 0);
        }

        // Open the FX window:
        else if (tr == GetLastTouchedTrack())
        {
          Main_OnCommand(ID_SHOW_FX_BROWSER, 0);
        }
      }
    }

    // Function 1 button?
    else if (evt->midi_message[1] == 0x36)
    {
      // Zoom:
      Main_OnCommand(ID_ZOOM_OUT_HORZ, 0);
    }

    // Function 2 button?
    else if (evt->midi_message[1] == 0x37)
    {
      // Zoom:
      Main_OnCommand(ID_ZOOM_IN_HORZ, 0);
    }

    // Function 3 button?
    else if (evt->midi_message[1] == 0x38)
    {
      // Zoom:
      Main_OnCommand(ID_ZOOM_OUT_VERT, 0);
    }

    // Function 4 button?
    else if (evt->midi_message[1] == 0x39)
    {
      // Zoom:
      Main_OnCommand(ID_ZOOM_IN_VERT, 0);
    }
  }

  // Note off?
  else if (evt->midi_message[0] == 0x80 || (evt->midi_message[0] == 0x90 && evt->midi_message[2] == 0))
  {
    // Forward button?
    if (evt->midi_message[1] == 0x5C)
    {
      // Reset button state:
      forwarding = false;
    }

    // Rewind button?
    else if (evt->midi_message[1] == 0x5B)
    {
      // Reset button state:
      rewinding = false;
    }

    // Touched fader?
    else if (evt->midi_message[1] == 0x68)
      touching = false;

    // EQ All Bypass?
    else if (evt->midi_message[1] == 0x75)
    {
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
      {
        // Check EQ:
        int fxcnt = TrackFX_GetCount(tr);
        if (fxcnt)
        {
          for (int fxid = 0; fxid < fxcnt; fxid++)
          {
            char fxname[256];
            if (TrackFX_GetFXName(tr, fxid, fxname, sizeof(fxname) / sizeof(fxname[0])))
            {
              if (strcmp(fxname, "VST: ReaEQ (Cockos)") == 0)
              {
                int paramcnt = TrackFX_GetNumParams(tr, fxid);
                for (int paramid = 0; paramid < paramcnt; paramid++)
                {
                  char paramname[256];
                  if (TrackFX_GetParamName(tr, fxid, paramid, paramname, sizeof(paramname) / sizeof(paramname[0])))
                  {
                    if (strcmp(paramname, "Bypass") == 0)
                    {
                      double minval, maxval, val;
                      val = TrackFX_GetParam(tr, fxid, paramid, &minval, &maxval);
                      TrackFX_SetParam(tr, fxid, paramid, 1.0 - val);
                      SwitchLight(LightEQBypass, val < 0.5);
                      lastEQChkTime = timeGetTime();
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  // Control change?
  else if (evt->midi_message[0] == 0xB0)
  {
    // Jog Dial?
    if (evt->midi_message[1] == 0x3C)
    {
      // Jogging?
      if (jogDown)
      {
        if (evt->midi_message[2] < 0x40)
          CSurf_ScrubAmt(evt->midi_message[2] / jogSpeed);
        else
          CSurf_ScrubAmt((evt->midi_message[2] & ~0x40) / -jogSpeed);
      }

      // No, control mode:
      else
      {
        // Get cursor position as POINTS:
        POINT p32;
        POINTS p16;
        GetCursorPos(&p32);
        p16.x = static_cast<SHORT>(p32.x);
        p16.y = static_cast<SHORT>(p32.y);

        // Get amount:
        int adj = (evt->midi_message[2] & ~0x40) * 120;

        // Get direction (negative values start at 0x40):
        if (evt->midi_message[2] & 0x40)
          adj = -adj;

        // Send the message:
        PostMessage(hReaperWnd, WM_MOUSEWHEEL, MAKEWPARAM(0, adj), reinterpret_cast<LPARAM>(&p16));
      }
    }

    // Pan?
    else if (evt->midi_message[1] == 0x10)
    {
      // Get current track:
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
      {
        // Get amount of change (encoder clicks) * (change per click):
        double adj = (evt->midi_message[2] & ~0x40) * (options.rotarySpeed / 1000.0);

        // Get direction (negative values start at 0x40):
        if (evt->midi_message[2] & 0x40)
          adj = -adj;

        // Update pan value:
        CSurf_SetSurfacePan(tr, CSurf_OnPanChange(tr, adj, true), NULL);
      }
    }

    // Value encoder?
    else if (evt->midi_message[1] == 0x50)
    {
      // Get master track:
      MediaTrack* tr = CSurf_TrackFromID(0, options.useMCP);
      if (tr)
      {
        // Get amount of change (encoder clicks) * (change per click):
        double adj = (evt->midi_message[2] & ~0x40) * (options.rotarySpeed / 20.0);

        // Get direction (negative values start at 0x40):
        if (evt->midi_message[2] & 0x40)
          adj = -adj;

        // Update volume value:
        CSurf_SetSurfaceVolume(tr, CSurf_OnVolumeChange(tr, adj, true), NULL);
      }
    }

    // EQ encoder?
    else if (evt->midi_message[1] >= 0x20 && evt->midi_message[1] <= 0x43)
    {
      // Get current track:
      MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
      if (tr)
      {
        // Check EQ:
        int fxcnt = TrackFX_GetCount(tr);
        if (fxcnt)
        {
          for (int fxid = 0; fxid < fxcnt; fxid++)
          {
            char fxname[256];
            if (TrackFX_GetFXName(tr, fxid, fxname, sizeof(fxname) / sizeof(fxname[0])))
            {
              if (strcmp(fxname, "VST: ReaEQ (Cockos)") == 0)
              {
                int paramcnt = TrackFX_GetNumParams(tr, fxid);
                int func = evt->midi_message[1] & 0xF0;
                int chan = (evt->midi_message[1] & 0x0F);
                int param;
                double scale;

                if (func == 0x20) // Q
                {
                  param = 2 + (chan * 3);
                  scale = 4000.0;
                }
                else if (func == 0x30) // F
                {
                  param = 0 + (chan * 3);
                  scale = 6000.0;
                }
                else // Gain
                {
                  param = 1 + (chan * 3);
                  scale = 4000.0;
                }

                // Get amount of change (encoder clicks) * (change per click):
                double adj = (evt->midi_message[2] & ~0x40) * (options.rotarySpeed / scale);

                // Get direction (negative values start at 0x40):
                if (evt->midi_message[2] & 0x40)
                  adj = -adj;

                //
                double minval, maxval, val;
                val = TrackFX_GetParam(tr, fxid, param, &minval, &maxval);
                if (func == 0x40)
                  val = volToInt14(val);
                val += adj;
                if (func == 0x40)
                  val = int14ToVol((int)val);
                if (val < minval)
                  val = minval;
                else if (val > maxval)
                  val = maxval;
                
                TrackFX_SetParam(tr, fxid, param, val + adj);

                break;
              }
            }
          }
        }
      }
    }
  }

  // Pitch bend? Must be the fader:
  if (evt->midi_message[0] == 0xE0)
  {
    MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
    if (tr)
    {
      CSurf_SetSurfaceVolume(tr, CSurf_OnVolumeChange(tr, int14ToVol(evt->midi_message[2], evt->midi_message[1]), false), NULL);
    }
  } 
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SkipTracks()
///////////////////////////////////////////////////////////////////////////////
// Description:  Internal helper function to move from track to track.
// Parameters:   [in] amt:   Amount of tracks to move, may be negative.
//               [in] dosel: Select the choosen track.
// Return value: No return value.
// Remarks:      The wrap around mode is controlled by the options.
///////////////////////////////////////////////////////////////////////////////
void CC121::SkipTracks(int amt, bool dosel)
{
  // Anything to do?
  if (amt == 0)
    return;

  // Calc new track ID:
  int newTrackID = currentTrackID + amt;

  // Clip value:
  if (newTrackID < 0)
  {
    if (options.wrapAround)
      newTrackID = CSurf_NumTracks(options.useMCP);
    else
      newTrackID = 0;
  }
  else if (amt > 0)
  {
    int trackCnt = CSurf_NumTracks(options.useMCP);
    if (newTrackID > trackCnt)
    {
      if (options.wrapAround)
        newTrackID = 0;
      else
        newTrackID = trackCnt;
    }
  }

  // Anything to do?
  if (newTrackID == currentTrackID)
    return;

  // Set new track ID:
  currentTrackID = newTrackID;

  // Force a channel state check:
  lastMonChkTime = 0;
  lastEQChkTime  = 0;
  lastFXChkTime  = 0;

  // Select it?
  if (dosel)
  {
    // Select our track:
    MediaTrack* tr = CSurf_TrackFromID(currentTrackID, options.useMCP);
    if (tr)
    {
      CSurf_SetSurfaceSelected(tr, CSurf_OnSelectedChange(tr, true), NULL);
      SetLastTouchedTrack(tr);
    }

    // Deselect all tracks:
    for (int x = 0; ; x++)
    {
      // Get track state:
      int flags = 0;
      if (!GetTrackInfo(x - 1, &flags))
        break;
      
      // Get track:
      MediaTrack* tt = CSurf_TrackFromID(x, options.useMCP);
      if (!tt || tr == tt)
        continue;

      // Skip already deselected tracks:
      if ((flags & 2) == 0)
        continue;

      // Deselect track:
      SetTrackSelected(tt, false);
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121::int14ToVol()
///////////////////////////////////////////////////////////////////////////////
// Description:  Convert a 14 bit value to a REAPER compatible volume value.
// Parameters:   [in] msb: Most significant bit of the 14 bit value.
//               [in] lsb: Least significant bit of the 14 bit value.
// Return value: Returns the converted volume.
// Remarks:      The volume value reaches from -oo to +24dB. This version
//               differs from the original SDK version as it introduces an
//               offset to hit the 0 dB marker on the actual control surface.
///////////////////////////////////////////////////////////////////////////////
double CC121::int14ToVol(unsigned char msb, unsigned char lsb)
{
  return int14ToVol(lsb | (msb << 7));
}

///////////////////////////////////////////////////////////////////////////////
// CC121::int14ToVol()
///////////////////////////////////////////////////////////////////////////////
// Description:  Convert a 14 bit value to a REAPER compatible volume value.
// Parameters:   [in] val: The 14 bit value.
// Return value: Returns the converted volume.
// Remarks:      The volume value reaches from -oo to +24dB. This version
//               differs from the original SDK version as it introduces an
//               offset to hit the 0 dB marker on the actual control surface.
///////////////////////////////////////////////////////////////////////////////
double CC121::int14ToVol(int val)
{
  double pos = (static_cast<double>(val) * 1000.0) / 16384.0;
  pos = SLIDER2DB(pos);
  return DB2VAL(pos - 0.33);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::volToInt14()
///////////////////////////////////////////////////////////////////////////////
// Description:  Convert a REAPER volume value to a 14 bit MSB/LSB value.
// Parameters:   [in] vol: The volume value.
// Return value: Returns the converted volume.
// Remarks:      The volume value reaches from -oo to +24dB. This version
//               differs from the original SDK version as it introduces an
//               offset to hit the 0 dB marker on the actual control surface.
///////////////////////////////////////////////////////////////////////////////
int CC121::volToInt14(double vol)
{
  double d = (DB2SLIDER(VAL2DB(vol) + 0.33) * 16384.0 / 1000.0);
  if (d <0.0)
    d = 0.0;
  else if (d > 16383.0)
    d = 16383.0;
  return static_cast<int>(d + 0.5);
}

///////////////////////////////////////////////////////////////////////////////
// CC121::SetLastTouchedTrack()
///////////////////////////////////////////////////////////////////////////////
// Description:  Internal helper to set the active track for REAPER's actions.
// Parameters:   [in] tr: The target track.
// Return value: No return value.
// Remarks:      Many of REPEAR's actions which are invoked with Main_OnCommand
//               work on the track that has been touched last by the user and
//               not the selected tracks.
///////////////////////////////////////////////////////////////////////////////
void CC121::SetLastTouchedTrack(MediaTrack* tr) const
{
  // Get positions of the current touched track and the desired track:
  int new_id = CSurf_TrackToID(tr, options.useMCP);
  int old_id = CSurf_TrackToID(GetLastTouchedTrack(), options.useMCP);

  // Move 'touch cursor' to the desired track (one track a time):
  while (old_id < new_id)
  {
    old_id++;
    Main_OnCommand(ID_TRACK_NEXT, 0);
  }
  while (old_id > new_id)
  {
		old_id--;
    Main_OnCommand(ID_TRACK_PREV, 0);
  }
}

///////////////////////////////// End of File /////////////////////////////////
