///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// File:           CC121Options.cpp                                          //
//                                                                           //
// Functionality:  Control surface option helper class implementation.       //
//                                                                           //
// Author:         Rolf Meyerhoff (rm@matrix44.de)                           //
//                                                                           //
// Copyright:      (C) 2009 by Rolf Meyerhoff. All rights reserved.          //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// License:                                                                  //
//                                                                           //
// This library is free software;  you can redistribute it  and/or modify it //
// under the terms of the GNU Lesser  General Public License as published by //
// the Free Software Foundation;  either version 2.1 of the License,  or (at //
// your option) any later version.                                           //
//                                                                           //
// This  library is  distributed in  the hope that  it will  be useful,  but //
// WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty   of //
// MERCHANTABILITY or FITNESS FOR A  PARTICULAR PURPOSE.  See the GNU Lesser //
// General Public License for more details.                                  //
//                                                                           //
// You should have received a copy of  the GNU Lesser General Public License //
// along with this library;  if not,  write to the Free Software Foundation, //
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.              //
///////////////////////////////////////////////////////////////////////////////
#include "reaper_cc121.h"

///////////////////////////////////////////////////////////////////////////////
// CC121Options::CC121Options()
///////////////////////////////////////////////////////////////////////////////
// Description:  Default constructor of this struct.
// Parameters:   None.
// Return value: No return value.
// Remarks:      Just initializes the members of this struct.
///////////////////////////////////////////////////////////////////////////////
CC121Options::CC121Options() :
  skipToStartEnd(false),
  rotarySpeed(31),
  midiInDev(-1),
  midiOutDev(-1),
  wrapAround(false),
  useMCP(false)
{
  // Nothing to do here.
}

///////////////////////////////////////////////////////////////////////////////
// CC121Options::CC121Options()
///////////////////////////////////////////////////////////////////////////////
// Description:  Copy constructor of this struct.
// Parameters:   [in] other: The source struct to copy.
// Return value: No return value.
// Remarks:      Just copies the members of the source struct.
///////////////////////////////////////////////////////////////////////////////
CC121Options::CC121Options(const CC121Options& other) :
  skipToStartEnd(other.skipToStartEnd),
  rotarySpeed(other.rotarySpeed),
  midiInDev(other.midiInDev),
  midiOutDev(other.midiOutDev),
  wrapAround(other.wrapAround),
  useMCP(other.useMCP)
{
  // Nothing to do here.
}

///////////////////////////////////////////////////////////////////////////////
// CC121Options::CC121Options()
///////////////////////////////////////////////////////////////////////////////
// Description:  Initialization constructor of this struct.
// Parameters:   [in] config: The members of this struct as text.
// Return value: No return value.
// Remarks:      The configuration is just a sequence of int values. This
//               constructor just loads these values and initializes the
//               members of this struct.
///////////////////////////////////////////////////////////////////////////////
CC121Options::CC121Options(const char* config) :
  skipToStartEnd(false),
  rotarySpeed(31),
  midiInDev(-1),
  midiOutDev(-1),
  wrapAround(false),
  useMCP(false)
{
  // Read configuration:
  const char* ptr = config;
  if (ptr)
  {
    // Loop through parameters:
    int param = 0;
    while (param < 6)
    {
      // skip leading spaces:
      while (*ptr == ' ')
        ptr++;

      // Is a number?
      if ((*ptr < '0' || *ptr > '9') && *ptr != '-')
        break;

      // Convert value to int:
      int val = atoi(ptr);

      // Fill parameter:
      switch (param)
      {
      case 0:
        skipToStartEnd = val != 0;
        break;
      case 1:
        rotarySpeed = val;
        break;
      case 2:
        midiInDev = val;
        break;
      case 3:
        midiOutDev = val;
        break;
      case 4:
        wrapAround = val != 0;
        break;
      case 5:
        useMCP = val != 0;
        break;
      }

      // Next parameter:
      param++;

      // Advance to next token:
      while (*ptr && *ptr != ' ')
        ptr++;
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// CC121Options::WriteConfig()
///////////////////////////////////////////////////////////////////////////////
// Description:  Write members of this struct to a string.
// Parameters:   [out] config: Receives the members of this struct as text.
// Return value: No return value.
// Remarks:      The configuration is just a sequence of int values. This
//               functions just writes the members of this struct into the
//               provided buffer.
///////////////////////////////////////////////////////////////////////////////
void CC121Options::WriteConfig(char* config) const
{
  // Parameter check:
  if (!config)
    return;

  // Fill buffer:
  sprintf(config, "%d %d %d %d %d %d", skipToStartEnd, rotarySpeed, midiInDev, midiOutDev, wrapAround, useMCP);
}

static INT_PTR CALLBACK dlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  // Initialization of the dialog, lParam points to the configuration string:
  if (uMsg == WM_INITDIALOG)
  {
    CC121Options opt((const char*)lParam);

    // Init skip markers check box:
    CheckDlgButton(hwndDlg, IDC_CHK_SKIP_MARKERS, opt.skipToStartEnd);

    // Init rotary speed:
    TCHAR buff[512];
    sprintf(buff, "%d", opt.rotarySpeed);
    SetWindowText(GetDlgItem(hwndDlg, IDC_EDIT_ENC_SPEED), buff);

    // Update spin control bounds and value:
    SendDlgItemMessage(hwndDlg, IDC_SPIN_ENC_SPEED, UDM_SETRANGE32, 1, 99);
    SendDlgItemMessage(hwndDlg, IDC_SPIN_ENC_SPEED, UDM_SETPOS32, 0, opt.rotarySpeed);

    // Add an 'unassigned' entry to the MIDI input list:
    LRESULT itemPos = SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_ADDSTRING, 0, (LPARAM)_T("None"));
    SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_SETITEMDATA, itemPos, -1);
    SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_SETCURSEL, itemPos, 0);

    // Loop through MIDI inputs:
    int numDevs = GetNumMIDIInputs();
    for (int i = 0; i < numDevs; i++)
    {
      // Get name of the device:
      if (GetMIDIInputName(i, buff, sizeof(buff) / sizeof(buff[0])))
      {
        // Add item to the combo box:
        itemPos = SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_ADDSTRING, 0, (LPARAM)buff);
        SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_SETITEMDATA, itemPos, i);

        // Select the entry if it represents the current device:
        if (i == opt.midiInDev)
          SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_SETCURSEL, itemPos, 0);
      }
    }

    // Add an 'unassigned' entry to the MIDI output list:
    itemPos = SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT, CB_ADDSTRING, 0, (LPARAM)"None");
    SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT, CB_SETITEMDATA, itemPos, -1);
    SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT, CB_SETCURSEL, itemPos, 0);

    // Loop through MIDI outputs:
    numDevs = GetNumMIDIOutputs();
    for (int i = 0; i < numDevs; i ++)
    {
      // Get name of the device:
      if (GetMIDIOutputName(i, buff, sizeof(buff) / sizeof(buff[0])))
      {
        // Add item to the combo box:
        itemPos = SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT, CB_ADDSTRING, 0, (LPARAM)buff);
        SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT, CB_SETITEMDATA, itemPos, i);

        // Select the entry if it represents the current device:
        if (i == opt.midiOutDev)
          SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT, CB_SETCURSEL, itemPos, 0);
      }
    }

    // Init wrap around and mcp check boxes:
    CheckDlgButton(hwndDlg, IDC_CHK_WRAP_AROUND, opt.wrapAround);
    CheckDlgButton(hwndDlg, IDC_CHK_USE_MCP,     opt.useMCP);
  }

  // Dialog item handling:
  else if (uMsg == WM_COMMAND)
  {
    // Check contents of the speed edit box:
    if (HIWORD(wParam) == EN_KILLFOCUS && LOWORD(wParam) == IDC_EDIT_ENC_SPEED)
    {
      // Get the text:
      TCHAR buff[256];
      if (SendMessage((HWND)lParam, WM_GETTEXT, sizeof(buff) / sizeof(buff[0]), (LPARAM)buff) > 0)
      {
        // Get value:
        int val = _tstoi(buff);

        // Update display on error:
        if (val < 1)
          SetWindowText((HWND)lParam, _T("1"));
        else if (val > 99)
          SetWindowText((HWND)lParam, _T("99"));
      }

      // Text retrieval was not successfull:
      else
        SetWindowText((HWND)lParam, _T("31"));
    }
  }

  // User message, request configuration where wParam is the size of a text
  // buffer pointed to by lParam.
  else if (uMsg == (WM_USER + 1024))
  {
    // Parameter check:
    if (wParam <= 1 || lParam == NULL)
      return 0;

    // Option set:
    CC121Options opt;

    // Skip markers:
    opt.skipToStartEnd = IsDlgButtonChecked(hwndDlg, IDC_CHK_SKIP_MARKERS) != 0;

    // Encoder speed:
    TCHAR buff[256];
    if (SendMessage(GetDlgItem(hwndDlg, IDC_EDIT_ENC_SPEED), WM_GETTEXT, sizeof(buff) / sizeof(buff[0]), (LPARAM)buff) > 0)
    {
      // Get value:
      opt.rotarySpeed = _tstoi(buff);

      // Check bounds:
      if (opt.rotarySpeed < 1)
        opt.rotarySpeed = 1;
      else if (opt.rotarySpeed > 99)
        opt.rotarySpeed = 99;
    }

    // MIDI in device:
    LRESULT itemPos = SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_GETCURSEL, 0, 0);
    if (itemPos != CB_ERR)
      opt.midiInDev = (int)SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_INPUT, CB_GETITEMDATA, itemPos, 0);

    // MIDI out device:
    itemPos = SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT,CB_GETCURSEL, 0, 0);
    if (itemPos != CB_ERR)
      opt.midiOutDev = (int)SendDlgItemMessage(hwndDlg, IDC_CB_MIDI_OUTPUT, CB_GETITEMDATA, itemPos, 0);

    // Wrap around and mcp check boxes:
    opt.useMCP     = IsDlgButtonChecked(hwndDlg, IDC_CHK_USE_MCP)     != 0;
    opt.wrapAround = IsDlgButtonChecked(hwndDlg, IDC_CHK_WRAP_AROUND) != 0;

    // Create configuration string:
    opt.WriteConfig((char*)lParam);
  }

  // Done:
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
// CreateConfigDialog()
///////////////////////////////////////////////////////////////////////////////
// Description:  Crete a dialog to edit the options of this extension.
// Parameters:   [in] type_string:      The type of surface to configure, this
//                                      is always "CC121" in our case.
//               [in] parent:           Parent window of the dialog.
//               [in] initConfigString: The current configuration.
// Return value: Returns a handle to the newly created window.
// Remarks:      This dialog is embedded into the REAPER configuration dialog.
///////////////////////////////////////////////////////////////////////////////
HWND CreateConfigDialog(const char* type_string, HWND parent, const char* initConfigString)
{
  // Create dialog from resource:
  return CreateDialogParam(hResourceInst, MAKEINTRESOURCE(IDD_SURFACEEDIT_CC121), parent, dlgProc, (LPARAM)initConfigString);
}

///////////////////////////////// End of File /////////////////////////////////
